/*
	ProWesS, PROGS Windowing System

	Directory Select type "DSEL"

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 5 September 1995
	23.04.1997 window now uses PW_OBJECT_AUTOSIZE for menu
*/

#include <mem.h>
#include <dir.h>
#include <str.h>
#include <dev.h>
#include <ctype.h>

#include "PWHandler.h"

#define TYPE PW_TYPE_DIRSELECT
#define catch(X) do { if (err=(X)) return err; } while (0)

/* global data */
typedef struct _List {
	struct _List *next;
	char name[2];
} List;

List devices	={NULL,""};
List directories={NULL,""};

/* loose item object definition */

typedef struct {
	PWBaseObject;
	PWObject outl, dir, menu;
	Error (*After)(PWObject obj, char *dir);	/* routine to process dir */
} Object;

/* object handling routines */

Error Create(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	/* allocate specific object - which should be cleared (automatic) */
	return PWAllocate(event.region->system,sizeof(Object),ret);
}

Error Remove(PWObject object)
{
	Object *obj=(Object *)object;

	if (obj->outl)
		PWRemove(obj->outl);
	return PWRelease((PWObject)obj,(PWObject)obj);
}

/* actions for directory select window */

Error getsubdir(PWObject object)
{
	Object *obj;
	DirIterator it;
	char currdir[IO_MAXDIRECTORY];
	char name[IO_MAXDIRECTORY+2];
	FileInfo fi;
	List *list;

	PWQuery(object,PW_GLOBAL_AUXILIARY,&obj);
	PWChange(obj->dir,PW_EDLINE_GET,IO_MAXDIRECTORY,currdir,NULL);
	PWChange(obj->menu,PW_MENU_CLEAR,NULL);

	if (*currdir=='\0')
		DEVDataGet(currdir);

	/* get subdirectories */
	if (!DIRSubOpen(currdir,FALSE,&it))
	{
		DIRSubDirectory(it,currdir);
		PWChange(obj->dir,PW_EDLINE_SET,currdir,NULL);
		while(!DIRSubNext(it,&fi))
		{
			STRCopy("> ",name);
			STRAppend(fi.name,name);
			PWChange(obj->menu,PW_MENU_ADD_COPY,name,NULL);
		}
		DIRSubClose(it);
	}

	/* add some auxiliary default devices etc. */
	for (list=directories.next ; list ; list=list->next)
	{
		PWChange(obj->menu,PW_MENU_ADD,list->name,NULL);
	}
	return ERR_OK;
}

Error ddirup(PWObject object)
{
	Object *obj;
	char currdir[IO_MAXDIRECTORY];
	int len;

	PWQuery(object,PW_GLOBAL_AUXILIARY,&obj);
	PWChange(obj->dir,PW_EDLINE_GET,IO_MAXDIRECTORY,currdir,NULL);

	len=strlen(currdir);
	for (len=strlen(currdir)-1 ; len>0 && currdir[len-1]!='_' ; len--);
	if (len>0)
	{
		currdir[len]='\0';
		PWChange(obj->dir,PW_EDLINE_SET,currdir,NULL);
		getsubdir(object);
	}
	return ERR_OK;
}

Error ddirdown(PWObject dir, char *item)
{
	char currdir[IO_MAXDIRECTORY];

	PWChange(dir,PW_EDLINE_GET,IO_MAXDIRECTORY,currdir,NULL);

	if (currdir[strlen(currdir)-1]!='_')
		STRAppendSafe("_",currdir,IO_MAXDIRECTORY);
	STRAppendSafe(item,currdir,IO_MAXDIRECTORY);
	PWChange(dir,PW_EDLINE_SET,currdir,NULL);

	return ERR_OK;
}

Error dsel(PWObject object, char *item)
{
	Object *obj;
	PWQuery(object,PW_GLOBAL_AUXILIARY,&obj);

	if (*item=='>')
		ddirdown(obj->dir,item+2);
	else
		PWChange(obj->dir,PW_EDLINE_SET,item,NULL);
	getsubdir(object);
}

Error devsel(PWObject object)
{
	Object *obj;
	char olddir_[IO_MAXDIRECTORY];
	char newdir[IO_MAXDIRECTORY];
	char *item, *pos, *olddir=olddir_;

	/* the item has to contain the underscore in the devicename !! */

	PWQuery(object,PW_GLOBAL_AUXILIARY,&obj);
	PWChange(obj->dir,PW_EDLINE_GET,IO_MAXDIRECTORY,olddir,NULL);
	PWQuery(object,PW_LOOSE_TEXT,&item);

	/* forget network node number !! */
	if (isalpha(olddir[0]) && isdigit(olddir[1]) && olddir[2]=='_')
		olddir+=3;

	if (MEMSame(strlen(item),item,olddir))
		return ERR_OK;

	for (pos=olddir ; *pos && *pos++!='_' ; );
	STRCopySafe(item,newdir,IO_MAXDIRECTORY);
	STRAppendSafe(pos,newdir,IO_MAXDIRECTORY);
	PWChange(obj->dir,PW_EDLINE_SET,newdir,NULL);
	getsubdir(object);

	return ERR_OK;
}

Error Handle(PWObject object, PWWait event)
{
	Object *obj=(Object *)object;
	char dir[IO_MAXDIRECTORY];
	Error err;

	getsubdir(obj->outl);
	PWChange(obj->outl,PW_WINDOW_ORIGIN_RESET,NULL);
	err=PWActivate(obj->outl);
	if (err) return err;
	PWChange(obj->dir,PW_EDLINE_GET,IO_MAXDIRECTORY,dir,NULL);
	if (obj->After)
		err=obj->After(object,dir);
	return err;
}

Error exit(PWObject object)
{
	return PWChange(object,PW_SYSTEM_BREAKDOWN,NULL);
}

Error Change(PWObject object, int *taglist)
{
	Object *obj=(Object *)object;
	PWSystem *system=obj->event.region->system;
	Error err=ERR_OK;
	char unrecognised_tag=FALSE;

	if (!(obj->status&OSTATUS_INITIALISED))
	{
		PWObject dirbox, devbox;
		List *dev;

		obj->status|=OSTATUS_INITIALISED;

		/* create all the objects for the directory select window */
		catch( PWCreate(NULL,&obj->outl,PW_TYPE_OUTLINE,
						PW_OUTLINE_QUIT,
						PW_OUTLINE_ACTION_WAKE, &getsubdir,
						PW_OUTLINE_QUIT_KEYPRESS, 27,
						PW_GLOBAL_AUXILIARY, obj,
						PW_OUTLINE_TITLE_TEXT, "dir select",
						NULL) );
		catch( PWCreate(obj->outl,NULL,PW_TYPE_KEYPRESS,
						PW_KEYPRESS, KEYSTROKE_DO,
						PW_KEYPRESS_ACTION, exit,
						NULL) );
		/* create edline for directory, with label */
		catch( PWCreate(obj->outl,&dirbox,PW_TYPE_DIRECTION,
						PW_SCALE_FACTOR, 0,
						NULL) );
		catch( PWCreate(dirbox,NULL,PW_TYPE_LABEL,
						PW_LABEL_TEXT, "directory",
						PW_LABEL_UNDERLINE, 'd',
						PW_SCALE_FACTOR, 0,
						NULL) );
		catch( PWCreate(dirbox,&obj->dir,PW_TYPE_EDLINE,
						PW_EDLINE_MAXLENGTH, IO_MAXDIRECTORY,
						PW_EDLINE_KEYPRESS, 'd',
						PW_EDLINE_ACTION_AFTER, &getsubdir,
						PW_EDLINE_WIDTH_FS, short2pt(12),
						NULL) );
		catch( PWCreate(dirbox,NULL,PW_TYPE_LOOSE_ITEM,
						PW_LOOSE_CHANGE_STATUS,FALSE,
						PW_LOOSE_ACTION_HIT, &ddirup,
						PW_LOOSE_TEXT, "<-",
						PW_LOOSE_KEYPRESS, '<',
						PW_SCALE_FACTOR, 0,
						PW_LOOSE_WINDOW_DO, TRUE,
						NULL) );
		/* create box for some devices */
		catch( PWCreate(obj->outl,&devbox,PW_TYPE_DIRECTION,
						PW_POSITION_BELOW, dirbox,
						PW_SCALE_FACTOR, 0,
						NULL) );
		/* create menu with possible filenames */
		catch( PWCreate(obj->outl,NULL,PW_TYPE_SEPARATOR,NULL) );
		catch( PWCreate(obj->outl,&obj->menu,PW_TYPE_MENU,
						PW_OBJECT_AUTOSIZE,
						PW_MENU_ACTION_SELECT, &dsel,
						PW_MENU_SORT_COMPARE, &STRCompareCI,
						PW_MENU_WINDOW_DO, TRUE,
						NULL) );
		/* create list of devices */
		for (dev=devices.next ; dev ; dev=dev->next)
		{
			catch( PWCreate(devbox,NULL,PW_TYPE_LOOSE_ITEM,
							PW_LOOSE_TEXT, dev->name,
							PW_LOOSE_CHANGE_STATUS, FALSE,
							PW_LOOSE_ACTION_HIT, &devsel,
							PW_LOOSE_WINDOW_DO, TRUE,
							NULL) );
		}
	}

	while (*taglist)
	{
		switch (*taglist)
		{
		case PW_DIRSELECT_SET:
			PWChange(obj->dir,PW_EDLINE_SET,(char *)(*(taglist+1)),NULL);
			break;
		case PW_DIRSELECT_TITLE_TEXT:
			PWChange(obj->outl,PW_OUTLINE_TITLE_TEXT,(char *)(*(taglist+1)),NULL);
			break;
		case PW_DIRSELECT_ACTIVATE:
			Handle(object,0);
			break;
		case PW_DIRSELECT_ACTION:
			obj->After=(Error (*)(PWObject,char *))(*(taglist+1));
			break;
		case PW_DIRSELECT_GET:
			PWChange(obj->dir,PW_EDLINE_GET,(*(taglist+1)),(*(taglist+2)),NULL);
			break;
		default :
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}

Error Define(char *name, char *value)
{
	List *link, *new;
	Error err;

	if (STRSameCD(name,"DIRSELECT-DEVICE"))
	{
		link=&devices;
linkname:
		catch( MEMAllocate(sizeof(List)-1+strlen(value),&new) );
		STRCopy(value,new->name);
		while (link->next) link=link->next;
		link->next=new;
		return ERR_OK;
	}
	if (STRSameCD(name,"DIRSELECT-DIRECTORY"))
	{
		link=&directories;
		goto linkname;
	}
	return ERR_IPAR;
}

PWTypeDef init={
	PW_TYPEDEF_FLAG, TYPE, PW_TYPE_EVENT_KEYPRESS,
	&Create, &Change, NULL, &Remove, NULL, NULL, &Handle, &Define,
	NULL, NULL, NULL
};
