/*
	ProWesS, PROGS Windowing System

	Label item object "LABL"

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 26 February 1995
	15.10.1995 increased itemwidth by 2 pixels to prevent clipping
	22.03.1996 adjusted code for internal use of pixel values
	10.04.1996 use PWHandleDefine && use ProWesS defaults
	25.11.1996 item height now better
	07.12.1996 use PFPWidth/Show/FontBbox
	23.01.1997 PW_LABEL_UNDERLINE added
*/

#include <mem.h>
#include <ctype.h>
#include "PWHandler.h"

#define TYPE PW_TYPE_LABEL

/* global statusses */

#define p100 short2pt(100)
ColourRGB ink = { 0, p100, 0 }; char Kink=FALSE;

char font[PF_MAXFONTNAME]="";		/* name of PROforma font to use */
pt fontsize=0;						/* fontsize to use */
pt baseline=0;						/* font baseline position (at fontsize) */
pt height=0;						/* height of font (at fontsize) */
char recalculate=TRUE;

/* loose item object definition */

typedef struct {
	PWBaseObject;
	char *item; 					/* item identifier */
	short itemwidth;
	char key;						/* char which should be underlined */
} Object;

/* object handling routines */

Error Create(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	LinkPROforma();
	/* allocate specific object - which should be cleared (automatic) */
	return PWAllocate(event.region->system,sizeof(Object),ret);
}

Error Change(PWObject object, int *taglist)
{
	Object *obj=(Object *)object;
	PWSystem *system=obj->event.region->system;
	Error err=ERR_OK;
	char unrecognised_tag=FALSE;

	while (*taglist)
	{
		switch (*taglist)
		{
		case PW_LABEL_TEXT:
			obj->item=(char *)(*(taglist+1));
			system->status|=SSTATUS_WINCHANGE;
			break;
		case PW_LABEL_UNDERLINE:
			obj->key=(*(taglist+1));
			system->status|=SSTATUS_REDRAW;
			break;
		default:
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}

Error Remove(PWObject object)
{
	return PWRelease(object,object);
}

Error Draw(PWObject object)
{
	Object *obj=(Object *)object;
	PWRegion *reg=obj->event.region;
	Gstate gstate=reg->system->gstate;
	pix hit_xsiz=reg->hit_xend-reg->hit_xorg;
	pix hit_ysiz=reg->hit_yend-reg->hit_yorg;
	pt xorg,yorg;

	if (!gstate) return ERR_OK;

	PFPWindowSubSet(gstate,hit_xsiz,hit_ysiz,reg->hit_xorg,reg->hit_yorg);
	PFColourRGB(gstate,&ink);
	PFFontLoad(gstate,font);
	PFFontScale(gstate,fontsize);
	if (reg->level&1)
	{
		/* left of item */
		xorg=short2pt(hit_xsiz-obj->itemwidth-1);
		yorg=baseline;
	} else {
		/* above item */
		xorg=pt_one;
		yorg=short2pt(hit_ysiz-height)+baseline;
	}
	PFPMoveTo(gstate,xorg,yorg);
	PFPStringShow(gstate,obj->item);
	if (obj->key)
	{
		Error err;
		char string[80],*pos;
		short strwidth;
		char key=obj->key;

		/* copy string and terminate at selection keystroke */
		STRCopySafe(obj->item,string,80);
		err=MEMSearch(key,strlen(string),string,&pos);
		if (err==ERR_ITNF)
		{
			key=islower(key)?toupper(key):tolower(key);
			err=MEMSearch(key,strlen(string),string,&pos);
		}
		if (err==ERR_OK)
		{
			*pos='\0';
			PFPStringWidth(gstate,string,&strwidth);
			PFPMoveTo(gstate,xorg+short2pt(strwidth),yorg+pt_one);

			string[0]=key;
			string[1]=0;
			PFPStringWidth(gstate,string,&strwidth);
			PFPathStroked(gstate);
			PFLineWidth(gstate, 0);
			PFPLineR(gstate,short2pt(strwidth),0);
			PFPathDraw(gstate);
		}
	}
	PFWindowSubRestore(gstate);

	return ERR_OK;
}

Defines defines[]={
{ "INK-COLOUR", PW_DEFINE_COLOUR,   NULL,   &ink,       &Kink },
{ "FONT",       PW_DEFINE_FONT,     NULL,   font,       &recalculate },
{ "FONTSIZE",   PW_DEFINE_PT,       NULL,   &fontsize,  &recalculate },
{ 0 }
};

Error Define(char *name, char *value)
{
	return PWHandleDefine("LABEL-",defines,name,value);
}

Error DetermineSize(PWObject object)
{
	Object *obj=(Object *)object;
	PWRegion *region=obj->event.region;
	Gstate gstate=region->system->gstate;

	region->hit_off_lft=region->hit_off_rgt=
	region->hit_off_top=region->hit_off_bot=0;

	if (!gstate) return ERR_OK;

	if (!Kink) PWQuery(object,PW_DEFAULT_MIDDLEGROUND,&ink);
	if (recalculate)
	{
		PBbox bbox;
		if (!fontsize) PWQuery(object,PW_DEFAULT_FONTSIZE,&fontsize);
		if (!*font) PWQuery(object,PW_DEFAULT_FONT,font);
		PFFontLoad(gstate,font);
		PFFontScale(gstate,fontsize);
		PFPFontBbox(gstate,0,&bbox);
		baseline=short2pt(bbox.yorg);
		height=bbox.ysiz;
		recalculate=FALSE;
	}

	/* set itemwidth */
	PFFontLoad(gstate,font);
	PFFontScale(gstate,fontsize);
	PFPStringWidth(gstate,obj->item,&obj->itemwidth);

	region->tot_ysiz=short2pt(height);
	region->tot_xsiz=short2pt(obj->itemwidth+2);

	return ERR_OK;
}


PWTypeDef init={
	PW_TYPEDEF_FLAG, TYPE, PW_TYPE_EVENT_REGION,
	&Create, &Change, NULL, &Remove, NULL, NULL, NULL, &Define,
	&Draw, &DetermineSize, NULL
};
